/*
  Macro           ScreenLocker
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Windows GUI versions of TSE Pro v4.2 upwards
  Version         v1.1   11 Jan 2024

  This tool intends to lock access to Windows until a password is re-entered.

  It has a screensaver!


  DISCLAIMER

  It has been tested for a long time on Windows 10 Pro now, and no one has
  reported being able to bypass the lock-screen yet.
  That said, the tool might have flaws nonetheless, because it has only been
  tested by a limited number of people, none of whome are known hackers or
  security professionals, and I am not a professional Windows programmer.
  Use at your own risk!


  CHALLENGE

  Please let me know if you find a way to access Windows without invoking the
  screenlocker's burglary reaction.
  Mention the Windows version if it is not Windows 10: I have this vague memory
  of many years ago breaking the method that ScreenLocker is now using.
  The assumption is that autorun is disabled for CDs and USB-devices, which I
  think is the default nowadays.


  INSTALLATION

  Copy this file to TSE's "mac" folder and compile it, for instance by opening
  the file in TSE and applying the Macro -> Compile menu.


  USAGE

  Execute the tool in any of these ways:

    Internally from inside the editor:
      By entering its name in TSE's Macro -> Execute menu.
      By adding it to TSE's Potpourri menu and starting it from there.
      By assigning a key to ExecMacro('ScreenLocker') in your .ui file.

    Externally from the command line or a Windows shortcut.
      Typically with:
        g32 -eScreenLocker
      If called externally this way, ScreenLocker takes two extra actions:
        It unloads all other macros to avoid conflicts.
        When finished it abandons the editor.
      Tip: Assign a Windows <CtrlAlt key> combination to the Windows shortcut.

  At any time after starting it, doing nothing starts the screensaver.

  After start-up the screenlocker every time asks for a password to use.
  A password can consist of any sequence of TSE keys except F10, Backspace
  and Enter.

  F10 lets you configure:
    The reaction to a "burglary attempt".
  Configurable reactions are:
    Just a warning.
    A Windows screenlock.
    A Windows screen-lock followed by a Windows shutdown.
  The default reaction is a warning without consequenses, so not a real lock.

  Attempting to leave the screen or exceeding the allowed number of password
  attempts invokes the configured burglary reaction.


  CAVEAT

  If Windows itself locks its screen after a non-usage time-out, then that also
  invokes the screenlocker's configured burglary reaction.


  TODO
    MUST
    SHOULD
    - Make the number of unlock attempts configurable.
    - Make <Escape> escape when entering a password.
    COULD
    - Redesign configuration:
      - First  reaction: Only warn. / Lock Windows.
      - Second reaction: Nothing. / Shutdown Windows after N seconds.
        N will be configurable.
    - Add the screensaver Sammy supplied and possibly other screensavers.
    WONT


  HISTORY

  v1.0      13 Jan 2020
    Initial version to test the concept.

  v1.0.1    27 May 2020
    No new features: This is just a technical update to make it compatible
    with TSE Pro v4.41.38 upwards by using v4.41.38's new built-in Random()
    function instead of the programmed one.

  v1.1      11 Jan 2024
    New:
      When ScreenLocker is loaded from the command line or a Windows shortcut,
      for example with "g32 -eScreenLocker", then unload all other extensions,
      and abandon the editor after the user unlocks or a "burglary" attempt.
    Fixed:
      Another extension (HistCleaner) accessing a network drive, which can
      happen up to minutes after starting TSE, made the editor lose focus,
      which ScreenLocker saw as a burglary attempt.
      Now ScreenLocker temporarily unloads HistCleaner.
*/





/*

  T E C H N I C A L   B  A C K G R O U ND    I N F O


  COMPATIBILITY

  The tool needs the Windows GUI version of TSE v4.2 upwards, because it needs
  the _LOSING_FOCUS_ and _NONEDIT_LOSING_FOCUS_ hooks.

*/





// Compatability check

#ifdef LINUX
  This tool is not compatible with the LINUX variant of TSE.
#endif

#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE 4.2.
#endif

#ifdef EDITOR_VERSION
#else
   Editor Version is older than TSE 3.0. You need at least TSE 4.2.
#endif

#if EDITOR_VERSION < 4200h
   Editor Version is older than TSE 4.2. You need at least TSE 4.2.
#endif

// End of compatability check.










Keydef no_keys
  <CtrlShift J> Right()
end no_keys





// Global constants and semi-constants
integer IS_LOADED_FROM_DOS_CMD    = FALSE
#define LOADED_MACROS_ID             4
#define LOCK_ATTR                   10  // 10 = Color(bright green ON black)
string  MACRO_NAME [MAXSTRINGLEN] = ''
integer ORG_ID                    = 0
integer LEAKED_KEYS_ID            =  0
integer SABOTEURS_ID              =  0
#define SCREENSAVER_DELAY           90  // In theory in 18ths of a second.


// Global variables
integer burglary_reaction               = 0
integer incompatibles_untested          = TRUE
string  input_password   [MAXSTRINGLEN] = ''
string  msg              [MAXSTRINGLEN] = ''
integer number_of_burglary_reactions    = 0
string  old_window_title [MAXSTRINGLEN] = ''
string  password         [MAXSTRINGLEN] = ''
integer screensaver_remaining_delay     = SCREENSAVER_DELAY
integer screen_is_locked                = FALSE
integer seed                            = 0
integer unlocked_CursorAttr             = 0
integer unlocked_TextAttr               = 0
integer unlock_tries_left               = 3


proc to_beep_or_not_to_beep()
  if Query(Beep)
    Sound(1500,  500)
    Sound( 500, 1500)
  endif
end to_beep_or_not_to_beep


#if EDITOR_VERSION <= 4400h
  integer proc Random()
    integer lo, Hi, test
    Hi = seed  /  127773
    lo = seed mod 127773
    test = 16807 * lo - (2147483647 mod 16807) * Hi
    if test > 0
      seed = test
    else
      seed = test + 2147483647
    endif
    return(seed)
  end Random
#endif


proc matrix_screensaver()
  string  a  [MAXSTRINGLEN] = ''
  string  s  [MAXSTRINGLEN] = ''
  integer p                 = 0
  integer x                 = 0
  integer y                 = 0
  for y = Query(ScreenRows) downto 2
    GetStrAttrXY(1, y - 1, s,  a, MAXSTRINGLEN)
    PutStrAttrXY(1, y    , s, '', LOCK_ATTR)
  endfor
  for x = 1 to Query(ScreenCols)
    p = Random() mod 100
    if p >= 95
      if s[x] == ' '
        s[x] = Chr(Random() mod 94 + 33)
      else
        s[x] = ' '
      endif
    else
      if s[x] <> ' '
        if p >= 15
          s[x] = Chr(Random() mod 94 + 33)
        else
          s[x] = ' '
        endif
      endif
    endif
  endfor
  PutStrAttrXY(1, 1, s, '', LOCK_ATTR)
end matrix_screensaver


proc display_helpline()
  PutStrAttrXY(1, Query(ScreenRows),
               'F10 menu   Backspace, Enter   Any other TSE key can be in the password.',
               '', LOCK_ATTR)
end display_helpline


proc display_password()
  PutStrAttrXY(10, 10, Format('':Length(input_password)/2:'*'), '', LOCK_ATTR)
end display_password


proc display_message()
  SetWindowTitle(msg)
  PutStrAttrXY(1, 1, Format(msg : Query(ScreenCols) * -1), '', LOCK_ATTR)
end display_message


proc idle()
  if incompatibles_untested
    if IS_LOADED_FROM_DOS_CMD
      // Assume all other loaded macros are saboteurs and purge them all.
      // This might corrupt keydefs after ScreenLocker, but in this branch
      // we will abandon the editor anyway.

      GotoBufferId(LOADED_MACROS_ID)
      PushBlock()
      MarkLine(1, NumLines())

      GotoBufferId(SABOTEURS_ID)
      EmptyBuffer()
      CopyBlock()
      UnMarkBlock()
      BegFile()

      repeat
        if  CurrLineLen()
        and GetText(3, 1) <> Chr(0)
        and isMacroLoaded(GetText(13, Asc(GetText(12, 1))))
        and not EquiStr(GetText(13, Asc(GetText(12, 1))), MACRO_NAME)
          PurgeMacro(GetText(13, Asc(GetText(12, 1))))
        endif
      until not Down()

      EmptyBuffer() // Do not reload the unloaded extensions.
    else
      // Temporarily unload specific macros.

      GotoBufferId(SABOTEURS_ID)
      BegFile()

      repeat
        if CurrLineLen()
          if isMacroLoaded(Trim(GetText(1, MAXSTRINGLEN)))
            PurgeMacro(Trim(GetText(1, MAXSTRINGLEN)))
          else
            BegLine()
            KillToEol()
          endif
        endif
      until not Down()
    endif

    GotoBufferId(LEAKED_KEYS_ID)
    Enable(no_keys, _EXCLUSIVE_)
    incompatibles_untested = FALSE
  endif

  if screensaver_remaining_delay
    if screensaver_remaining_delay mod 18 == 0
      Set(Cursor    , OFF)
      Set(TextAttr  , LOCK_ATTR)
      Set(CursorAttr, LOCK_ATTR)
      Set(Attr      , LOCK_ATTR)
      ClrScr()
      display_message()
      display_password()
      if not screen_is_locked
        display_helpline()
      endif
    endif
    screensaver_remaining_delay = screensaver_remaining_delay - 1
  else
    matrix_screensaver()
  endif
end idle


proc react_to_burglary_attempt()
  number_of_burglary_reactions = number_of_burglary_reactions + 1
  if number_of_burglary_reactions == 1 // Once is enough.
    case burglary_reaction
      when 2
        Dos('%WINDIR%\System32\rundll32.exe user32.dll,LockWorkStation',
            _DONT_PROMPT_|_DONT_CLEAR_|_RUN_DETACHED_|_START_HIDDEN_)
      when 3
        // First lock the screen, then shutdown.
        // The quickest way to shutdown still requires at least a 1 second,
        // while locking the screen can be done immediately.
        Dos('%WINDIR%\System32\rundll32.exe user32.dll,LockWorkStation',
            _DONT_PROMPT_|_DONT_CLEAR_|_RUN_DETACHED_|_START_HIDDEN_)
        Dos('shutdown /s /t 1')
    endcase
    Sound(1000, 250)
    Sound(1200, 250)
    Sound(1000, 250)
    Sound( 800, 250)
    Warn   ('There was a burglary attempt!')
    Message('There was a burglary attempt!') // Overwrites any obsolete message.
    PurgeMacro(MACRO_NAME)
  endif
end react_to_burglary_attempt


proc set_configured_burglary_reaction(integer p_burglary_reaction)
  burglary_reaction = p_burglary_reaction
  if not WriteProfileInt(MACRO_NAME + ':Configuration',
                         'BurglaryReaction',
                         burglary_reaction)
    Warn('Failed to save configuration setting.')
  endif
end set_configured_burglary_reaction


integer proc get_configured_burglary_reaction()
  burglary_reaction = GetProfileInt(MACRO_NAME + ':Configuration',
                                    'BurglaryReaction',
                                    1)
  return(burglary_reaction)
end get_configured_burglary_reaction


menu configuration_menu()
  title   = 'Burglary reaction'
  history = get_configured_burglary_reaction()
  '&Warning message and sound',
    set_configured_burglary_reaction(1),,
    'Make a sound and show a warning message.'
  '&Lock Windows',
    set_configured_burglary_reaction(2),,
    'Lock Windows.'
  '&Shutdown Windows',
    set_configured_burglary_reaction(3),,
    'Shut Windows down.'
end configuration_menu


proc configure()
  UnHook(idle)
  // Set(TextAttr, unlocked_attr)
  configuration_menu()
  // Set(TextAttr, LOCK_ATTR)
  Hook(_IDLE_        , idle)
  Hook(_NONEDIT_IDLE_, idle)
end configure


proc losing_focus()
  Message('Attempt to leave the screen locker detected!')
  react_to_burglary_attempt()
end losing_focus


proc after_getkey()
  integer key_code = Query(Key)
  screensaver_remaining_delay = SCREENSAVER_DELAY
  if screen_is_locked
    msg = 'The screen is locked: Enter password plus <Enter> to unlock.'
    if (key_code in <Enter>, <GreyEnter>)
      if input_password == password
        UnHook(losing_focus)
        msg = 'The screen is unlocked.'
        PurgeMacro(MACRO_NAME)
      else
        unlock_tries_left = unlock_tries_left - 1
        if unlock_tries_left == 0
          msg = 'Too many password attempts!'
          Message(msg)
          react_to_burglary_attempt()
        else
          msg = Format('Wrong password: You have ', unlock_tries_left,
                       ' attempts left.')
          input_password = ''
        endif
      endif
    elseif key_code == <BackSpace>
      if Length(input_password)
        input_password = input_password[1 : Length(input_password) - 2]
      else
        to_beep_or_not_to_beep()
      endif
    else
      if Length(input_password) == 254
        to_beep_or_not_to_beep()
      else
        input_password = input_password        +
                         Chr(HiByte(key_code)) +
                         Chr(LoByte(key_code))
        display_password()
      endif
    endif
  else
    msg = 'Type a new password plus <Enter> to lock the screen.'
    if (key_code in <Enter>, <GreyEnter>)
      msg = 'The screen is locked: Enter password plus <Enter> to unlock.'
      password         = input_password
      input_password   = ''
      screen_is_locked = TRUE
    elseif key_code == <BackSpace>
      if Length(input_password)
        input_password = input_password[1 : Length(input_password) - 2]
      else
        to_beep_or_not_to_beep()
      endif
    elseif key_code == <F10>
      configure()
      msg = 'End of configuration. Type a new password plus <Enter> to lock the screen.'
      input_password = ''
    else
      if Length(input_password) == 254
        to_beep_or_not_to_beep()
      else
        input_password = input_password        +
                         Chr(HiByte(key_code)) +
                         Chr(LoByte(key_code))
      endif
      display_password()
    endif
  endif
  BreakHookChain() // Stop CUAMark and friends nicely from processing keys this way.
end after_getkey


proc WhenPurged()
  Disable(no_keys)

  // Do a hard ignore to avoid leaking the last key, usually <Enter>.
  Set(Key, -1)
  Set(Cursor, ON)
  Set(TextAttr  , unlocked_TextAttr  )
  Set(CursorAttr, unlocked_CursorAttr)

  if SABOTEURS_ID
    GotoBufferId(SABOTEURS_ID)
    BegFile()
    repeat
      if      CurrLineLen()
      and not isMacroLoaded(Trim(GetText(1, MAXSTRINGLEN)))
        LoadMacro(Trim(GetText(1, MAXSTRINGLEN)))
      endif
    until not Down()
  endif

  GotoBufferId(org_id)
  AbandonFile(SABOTEURS_ID)
  AbandonFile(LEAKED_KEYS_ID)

  SetWindowTitle(old_window_title)
  UpdateDisplay(_ALL_WINDOWS_REFRESH_)

  if IS_LOADED_FROM_DOS_CMD
    AbandonEditor()
  endif
end WhenPurged


proc WhenLoaded()
  MACRO_NAME          = SplitPath(CurrMacroFilename(), _NAME_)
  ORG_ID              = GetBufferId()
  old_window_title    = GetWindowTitle()
  seed                = GetClockTicks() mod 1000000000 - seed
  unlocked_CursorAttr = Query(CursorAttr)
  unlocked_TextAttr   = Query(TextAttr  )
  burglary_reaction   = get_configured_burglary_reaction()

  if FALSE
  endif

  if Pos('-e'  + Lower(MACRO_NAME), Lower(Query(DosCmdLine)))
  or Pos('-e ' + Lower(MACRO_NAME), Lower(Query(DosCmdLine)))
  or Pos('-l'  + Lower(MACRO_NAME), Lower(Query(DosCmdLine)))
  or Pos('-l ' + Lower(MACRO_NAME), Lower(Query(DosCmdLine)))
    IS_LOADED_FROM_DOS_CMD = TRUE
  endif

  if isGUI()
    SABOTEURS_ID   = CreateTempBuffer()

    // These macros are innocent:
    //  AddLine('Git')
    //  AddLine('Uncrash')
    //  AddLine('Settings')
    //  AddLine('Keys')
    //  AddLine('Compile')
    //  AddLine('KeepBlinking')
    //  AddLine('WhtSpc')
    //  AddLine('Book')
    //  AddLine('EolType')
    //  AddLine('fState')
    //  AddLine('MkSavDir')
    //  AddLine('SynCfg2')
    //  AddLine('Uniview')
    //  AddLine('SynCase')
    //  AddLine('PickList')
    //  AddLine('HistList')
    //  AddLine('ControlChars')
    //  AddLine('HiliteAll')
    //  AddLine('OnlyOneSplitLinesWarning')
    //  AddLine('Lengths')
    //  AddLine('ClipPrmt')
    //  AddLine('WinUnMark')
    //  AddLine('TabExpan')

    //  HistCleaner turns out to disrupt ScreenLocker
    //  when HistCleaner accesses a network drive for the first time.
    AddLine('HistCleaner')

    LEAKED_KEYS_ID = CreateTempBuffer()
    msg = 'Type a password plus <Enter> to lock the screen.'
    Hook(_AFTER_GETKEY_        , after_getkey)
    Hook(_IDLE_                , idle        )
    Hook(_LOSING_FOCUS_        , losing_focus)
    Hook(_NONEDIT_IDLE_        , idle        )
    Hook(_NONEDIT_LOSING_FOCUS_, losing_focus)
  else
    to_beep_or_not_to_beep()
    Warn(MACRO_NAME, ' only works in the GUI Version of TSE.')
    PurgeMacro(MACRO_NAME)
  endif
end WhenLoaded

